/* $Id: pac_parse.yy 1436 2005-10-01 00:45:24Z vern $ */

%token TOK_TYPE TOK_RECORD TOK_CASE TOK_OF TOK_ENUM TOK_DEFAULT
%token TOK_LET TOK_PADDING TOK_TO TOK_ALIGN
%token TOK_ATTR_BYTEORDER TOK_ATTR_CHECK TOK_ATTR_ENDWITH TOK_ATTR_LENGTH
%token TOK_INT8 TOK_INT16 TOK_INT32
%token TOK_UINT8 TOK_UINT16 TOK_UINT32
%token TOK_ID TOK_NUMBER

%nonassoc '='
%left ';'
%left ','
%left '?' ':'
%left TOK_OR
%left TOK_AND
%nonassoc TOK_EQUAL TOK_LE TOK_GE '<' '>'
%left '&' '|' '^'
%left TOK_LSHIFT TOK_RSHIFT 
%left '+' '-'
%left '*' '/' '%'
%right '~' '!'
%right TOK_SIZEOF TOK_OFFSETOF
%left '.'
%right '(' ')' '[' ']'

%type <decl> decl
%type <type> type type3 type2 type1 type0
%type <paramlist> optparams paramlist
%type <param> param
%type <exprlist> exprlist
%type <fieldlist> fieldlist
%type <field> field
%type <padding> padding
%type <caseindex> caseindex
%type <caselist> caselist
%type <casespec> case
%type <enumlist> enumlist enumlist1
%type <enumitem> enumitem
%type <expr> expr
%type <seqend> seqend
%type <numlist> numlist
%type <attrlist> optattrs attrlist
%type <attr> attr
%type <id> TOK_ID
%type <num> TOK_NUMBER

%{
#include "pac.h"

extern int yyerror(const char msg[]);
extern int yylex();
extern char* yytext;

extern const char* input_filename;
extern int line_number;
extern Output* header_output;
extern Output* source_output;

vector<Decl*> decls;

void process_decl(Decl* decl)
	{
	decl->genCode(header_output, source_output);
	decls.push_back(decl);
	}

%}

%union {
	Decl		*decl;
	Type 		*type;
	Param 		*param;
	Field 		*field;
	PaddingField 	*padding;
	CaseIndex	*caseindex;
	Case 		*casespec;
	SeqEnd		*seqend;
	Expr		*expr;
	Attr		*attr;
	Enum		*enumitem;
	ParamList 	*paramlist;
	FieldList 	*fieldlist;
	CaseList 	*caselist;
	ExprList 	*exprlist;
	EnumList	*enumlist;
	NumList		*numlist;
	AttrList	*attrlist;
	ID		*id;
	Number		*num;
	const char	*str;
}

%%

decls		:	/* empty */
				{
				// Put initialization here
				/*
				process_decl(new LetDecl(
					new ID("bigendian"),
					new Expr(new Number(0))));
				process_decl(new LetDecl(
					new ID("littleendian"),
					new Expr(new Number(1))));
				*/
				}
		|	decls  decl ';'
				{
				process_decl($2);
				}
		;

decl		:	TOK_TYPE TOK_ID { current_decl_id = $2; } optparams '=' type optattrs
				{
				TypeDecl* decl = new TypeDecl($2, $4, $6);
				decl->AddAttr($7);
				$$ = decl;
				}

		|	TOK_LET TOK_ID { current_decl_id = $2; } '=' expr
				{
				$$ = new LetDecl($2, $5);
				}
		|	TOK_ENUM TOK_ID { current_decl_id = $2; } '{' enumlist '}'
				{
				$$ = new EnumDecl($2, $5);
				}
		;

optparams	:	'(' paramlist ')'
				{
				$$ = $2;
				}
		|	/* empty */
				{
				$$ = 0;
				}
		;

paramlist	:	paramlist ',' param 
				{
				$1->push_back($3);
				$$ = $1;
				}
		|	param
				{
				$$ = new ParamList();
				$$->push_back($1);
				}
		;

param		:	TOK_ID
				{
				$$ = new Param($1);
				}
		|	TOK_ID ':' TOK_ID
				{
				$$ = new Param($1, $3);
				}
		;

type		: 	type3
				{
				$$ = $1;
				}
		;

/* type3 is for record or case or type2 */
type3		:	type2
				{
				$$ = $1;
				}
		|	TOK_RECORD '{' fieldlist '}'
				{
				$$ = new RecordType($3);
				}
		;

/* type2 is for array or sequence or type1 */
type2		:	type1
				{
				$$ = $1;
				}
		|	type1 '[' expr ']'
				{
				$$ = new ArrayType($1, $3);
				}
		|	type1 '[' ']' TOK_ATTR_ENDWITH seqend
				{
				$$ = new ArrayType($1, $5);
				}
		|	TOK_CASE caseindex TOK_OF '{' caselist '}'
				{
				$$ = new CaseType($2, $5);
				}
		;	

/* type1 is for parameterized types or type0 */
type1		:	type0
				{
				$$ = $1;
				}
		| 	TOK_ID
				{
				int basic_type = 
					BasicType::lookup_basictype($1->Name());
				if ( basic_type >= 0 ) 
					$$ = new BasicType((BasicType::basic_type) basic_type);
				else
					$$ = new ParameterizedType($1, 0);
				}
		|	TOK_ID '(' exprlist ')'
				{
				$$ = new ParameterizedType($1, $3);
				}
		;

/* type0 is for basic types */
type0		:	TOK_INT8
				{
				$$ = new BasicType(BasicType::INT8);
				}
		|	TOK_INT16
				{
				$$ = new BasicType(BasicType::INT16);
				}
		|	TOK_INT32
				{
				$$ = new BasicType(BasicType::INT32);
				}
		|	TOK_UINT8
				{
				$$ = new BasicType(BasicType::UINT8);
				}
		|	TOK_UINT16
				{
				$$ = new BasicType(BasicType::UINT16);
				}
		|	TOK_UINT32
				{
				$$ = new BasicType(BasicType::UINT32);
				}
		;

fieldlist	:	fieldlist field ';'
				{
				$1->push_back($2);
				$$ = $1;
				}
		|	/* empty */	
				{
				$$ = new FieldList();
				}
		;

field		:	TOK_ID ':' type2 optattrs
				{
				$$ = new RecordField($1, $3);
				$$->AddAttr($4);
				}
		|	TOK_LET TOK_ID '=' expr optattrs
				{
				$$ = new LetField($2, $4);
				$$->AddAttr($5);
				}
		|	padding
				{
				$$ = $1;
				}
		;

padding		:	TOK_ID ':' TOK_PADDING '[' expr ']'
				{
				$$ = new PaddingField($1,
					PAD_BY_LENGTH, $5);
				}
		|	TOK_ID ':' TOK_PADDING TOK_TO expr
				{
				$$ = new PaddingField($1, 
					PAD_TO_OFFSET, $5);
				}
		|	TOK_ID ':' TOK_PADDING TOK_ALIGN expr
				{
				$$ = new PaddingField($1, 
					PAD_TO_NEXT_WORD, $5);
				}
		;

caseindex	:	expr
				{
				$$ = new CaseIndex($1);
				}
		;

caselist	:	caselist case ';'
				{
				$1->push_back($2);
				$$ = $1;
				}
		|	/* empty */	
				{
				$$ = new CaseList();
				}
		;

case		:	expr TOK_GE TOK_ID ':' type2
				{
				$$ = new Case($1, $3, $5);
				}
		|	TOK_DEFAULT TOK_GE TOK_ID ':' type2
				{
				$$ = new Case(0, $3, $5);
				}
		;

exprlist	:	exprlist ',' expr
				{
				$1->push_back($3);
				$$ = $1;
				}
		|	expr
				{
				$$ = new ExprList();
				$$->push_back($1);
				}
		;

expr		:	TOK_ID
				{
				$$ = new Expr($1);
				}
		|	TOK_NUMBER
				{
				$$ = new Expr($1);
				}
		|	expr '[' expr ']'
				{
				$$ = new Expr(Expr::EXPR_SUBSCRIPT, $1, $3);
				}
		|	expr '.' TOK_ID
				{
				$$ = new Expr(Expr::EXPR_MEMBER, $1, new Expr($3));
				}
		|	TOK_SIZEOF '(' TOK_ID ')'
				{
				$$ = new Expr(Expr::EXPR_SIZEOF, new Expr($3));
				}
		|	TOK_OFFSETOF '(' TOK_ID ')'
				{
				$$ = new Expr(Expr::EXPR_OFFSETOF, new Expr($3));
				}
		|	'(' expr ')'
				{
				$$ = new Expr(Expr::EXPR_PAREN, $2);
				}
		|	'-' expr
				{
				$$ = new Expr(Expr::EXPR_NEG, $2);
				}
		|	expr '+' expr
				{
				$$ = new Expr(Expr::EXPR_PLUS, $1, $3);
				}
		|	expr '-' expr
				{
				$$ = new Expr(Expr::EXPR_MINUS, $1, $3);
				}
		|	expr '*' expr
				{
				$$ = new Expr(Expr::EXPR_TIMES, $1, $3);
				}
		|	expr '/' expr
				{
				$$ = new Expr(Expr::EXPR_DIV, $1, $3);
				}
		|	expr '%' expr
				{
				$$ = new Expr(Expr::EXPR_MOD, $1, $3);
				}
		|	'~' expr
				{
				$$ = new Expr(Expr::EXPR_BITNOT, $2);
				}
		|	expr '&' expr
				{
				$$ = new Expr(Expr::EXPR_BITAND, $1, $3);
				}
		|	expr '|' expr
				{
				$$ = new Expr(Expr::EXPR_BITOR, $1, $3);
				}
		|	expr '^' expr
				{
				$$ = new Expr(Expr::EXPR_BITXOR, $1, $3);
				}
		|	expr TOK_LSHIFT expr
				{
				$$ = new Expr(Expr::EXPR_LSHIFT, $1, $3);
				}
		|	expr TOK_RSHIFT expr
				{
				$$ = new Expr(Expr::EXPR_RSHIFT, $1, $3);
				}
		|	expr TOK_EQUAL expr
				{
				$$ = new Expr(Expr::EXPR_EQUAL, $1, $3);
				}
		|	expr TOK_GE expr
				{
				$$ = new Expr(Expr::EXPR_GE, $1, $3);
				}
		|	expr TOK_LE expr
				{
				$$ = new Expr(Expr::EXPR_LE, $1, $3);
				}
		|	expr '>' expr
				{
				$$ = new Expr(Expr::EXPR_GT, $1, $3);
				}
		|	expr '<' expr
				{
				$$ = new Expr(Expr::EXPR_LT, $1, $3);
				}
		|	'!' expr
				{
				$$ = new Expr(Expr::EXPR_NOT, $2);
				}
		|	expr TOK_AND expr
				{
				$$ = new Expr(Expr::EXPR_AND, $1, $3);
				}
		|	expr TOK_OR expr
				{
				$$ = new Expr(Expr::EXPR_OR, $1, $3);
				}
		|	expr '?' expr ':' expr
				{
				$$ = new Expr(Expr::EXPR_COND, $1, $3, $5);
				}
		;

seqend		:	'[' numlist ']'
				{
				$$ = new SeqEnd($2);
				}
		;

numlist		:	numlist ',' TOK_NUMBER
				{
				$1->push_back($3);
				$$ = $1;
				}
		|	TOK_NUMBER
				{
				$$ = new NumList();
				$$->push_back($1);
				}
		;

enumlist	: 	enumlist1
				{
				$$ = $1;
				}
		|	enumlist1 ','
				{
				$$ = $1;
				}
		;

enumlist1	:	enumlist1 ',' enumitem
				{
				$1->push_back($3);
				$$ = $1;
				}
		|	enumitem
				{
				$$ = new EnumList();
				$$->push_back($1);
				}
		;

enumitem	:	TOK_ID
				{
				$$ = new Enum($1);
				}
		|	TOK_ID '=' expr
				{
				$$ = new Enum($1, $3);
				}
		;

optattrs	:	/* empty */	
				{
				$$ = 0;
				}
		|	attrlist
				{
				$$ = $1;
				}
		;

attrlist	:	attrlist ',' attr
				{
				if ( $3 )
					$1->push_back($3);
				$$ = $1;
				}
		|	attr
				{
				$$ = new AttrList();
				if ( $1 )
					$$->push_back($1);
				}
		;

attr		:	TOK_ATTR_BYTEORDER '=' expr
				{
				$$ = new AttrByteOrder($3);
				}
		|	TOK_ATTR_CHECK '(' expr ')'
				{
				$$ = new AttrCheck($3);
				}
		|	TOK_ATTR_LENGTH '=' expr
				{
				// TODO: to flesh out &length attribute
				$$ = 0;
				}
		;

%%

const ID* current_decl_id = 0;

int yyerror(const char msg[])
	{
	char* msgbuf = new char[strlen(msg) + strlen(yytext) + 64];

	if ( yytext[0] == '\n' )
		sprintf(msgbuf, "%s, on previous line", msg);

	else if ( yytext[0] == '\0' )
		sprintf(msgbuf, "%s, at end of file", msg);

	else
		sprintf(msgbuf, "%s, at or near \"%s\"", msg, yytext);

	/*
	extern int column;
	sprintf(msgbuf, "%*s\n%*s\n", column, "^", column, msg);
	*/

	if ( input_filename )
		fprintf(stderr, "%s:%d: ", input_filename, line_number);
	else
		fprintf(stderr, "line %d: ", line_number);
	fprintf(stderr, "%s", msgbuf);
	fprintf(stderr, " (yychar=%d)", yychar);
	fprintf(stderr, "\n");

	return 0;
        }
